/***********************************************************************/
#include <strings.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include "filenames.h"
#include "process_queries.h"

int mkdir(char *, mode_t);
int chdir(char *);

/***********************************************************************/
void OpenFiles(char *, FILE **, FILE **, FILE **);
void ProcessInputFile(FILE *, FILE *, FILE *);
void CloseFiles(FILE *, FILE *, FILE *);
void ProcessQuery(char *, FILE *, FILE *);
void Panic(char *);
int GetNextWord(char **, char *);
void GenerateQueryWeightVector(char *, int, int *);
void DB_ID(char *, char *);
void OpenQueryFiles(char *, char *, char *, FILE **, FILE **, FILE **);
void CloseQueryFiles(FILE *, FILE *, FILE *, char *);
void ProcessQueryFiles(char *, long, FILE *, FILE *, FILE *, t_exp *);
int ProcessQueryData(int *, char *, long, FILE *, FILE *, FILE *, t_exp *);
int ReadQuery(FILE *, int *, int **);
void ReadAndCollapseResults(FILE *, int, int, t_qres *, t_exp *);
void ReadAndCollapseResultsForOne(FILE *, int, long *, float *, int, t_exp *);
float RealGoodness(int, t_qres *, t_exp *);
float GlOSSGoodness(int, t_qres *, long, t_exp *, int *);
int query_results_compare(void *, void *);
void OpenRankFiles(char *, char *, FILE **, FILE **);
void ComputeStats(int, FILE *, FILE *, t_exp *);
void CloseRankFiles(FILE *, FILE *);
void CreateVectors(int, float **, float **, float **, float *, float *);
void DestroyVectors(float *, float *, float *);
int ReadInOneQuery(int, FILE *, FILE *, float *, float *, float *);
void AnalyzeData(int, float *, float *, float *, float *, float *, t_exp *);
void ReportResults(int, int, float *, float *, t_exp *);

/***********************************************************************/

char gbuffer[MAXLINE]; /* global variable, keeping the last
     line that we read from a query result file, which we shouldn't
     have read. Ungetc only guarantees ungetting one char.
     Horrible patch. Temporary... */

long zero_real_goodness=0; /* global variable: keeps track of 
     how many (database, query) pairs had zero goodness... */ 

long zero_gloss_goodness=0; /* global variable: keeps track of 
     how many (database, query) pairs had zero goodness as predicted
     by GlOSS... */ 

long total_goodness=0; /* global variable: keeps track of how many
     (database, query) pairs there are, for checking...*/

/***********************************************************************/
void main(int argc, char *argv[]){
int num_dbs;
t_exp experiment; 

if(argc!=6){
  fprintf(stderr, "Usage %s: \n \tReal_Rank_Algorithm [1..2] \n\
\tGlOSS_Rank_Algorithm [3..6] \n \tThreshold_Real [0..1]\n\
\tThreshold_GlOSS [0..1]\n \tOutput_File (without path)\n", argv[0]);  
  exit(0);
} 

/* initialize global gbuffer */
gbuffer[0]='\0';

/*
PreprocessSiftQueries(SIFT_QUERIES);
GenerateSmartQueries(SIFT_QUERIES);
ProcessDBList(DB_LIST);
FeedSmart(PROC_DB_LIST, 0, 0, NTC_NNN_DIR);
*/

/* Set some experiment parameters */
experiment.real_rank_alg = atoi(argv[1]);
experiment.gloss_rank_alg= atoi(argv[2]);
sscanf(argv[3], "%f", &(experiment.threshold_real)); 
sscanf(argv[4], "%f", &(experiment.threshold_gloss)); 
sprintf(experiment.output_file, "%s/%s", NTC_NNN_DIR, argv[5]);
experiment.prune=PRUNED_RANKS;

/* Run experiments */

GenerateRanks(PROC_DB_LIST, SIFT_QUERIES, NTC_NNN_DIR, 
              &num_dbs, &experiment);

CompareRanks(num_dbs, &experiment);

}

/***********************************************************************/
void PreprocessSiftQueries(char *filename){
/* file filename.all contains lines like:

   Profile hello
   Type    0
   Profile goodbye
   Type    1

  These are SIFT profiles (see http://sift.stanford.edu).
  Type 0 indicates vector-space profiles.
  Type 1 indicates boolean profiles.
  So far, we're just interested in the vector-space profiles.
  These are moved to file filename.prs 
*/

char f_tmp1[FILENAME_MAX], f_tmp2[FILENAME_MAX];
FILE *tmp1, *tmp2;
char buffer1[MAXLINE], buffer2[MAXLINE];

sprintf(f_tmp1, "%s/%s.all", QUERY_DIR, filename);
sprintf(f_tmp2, "%s/%s.prs", QUERY_DIR, filename);

if((tmp1=fopen(f_tmp1, "r"))==(FILE *)NULL)
  Panic("Couldn't open SIFT-queries file");

if((tmp2=fopen(f_tmp2, "w"))==(FILE *)NULL)
  Panic("Couldn't create processed SIFT-queries file");

while(fgets(buffer1, MAXLINE, tmp1)!=NULL){
   fgets(buffer2, MAXLINE, tmp1);
   if(strstr(buffer2, "Type       0")!=NULL)
     /* vector-space query: save, stripping the word "Profile" off */
     fprintf(tmp2, "%s", buffer1+12);
   else{
     if(strstr(buffer2, "Type       1")==NULL)
       /* unknown type */
       Panic("Unknown SIFT query type");
   }
}

fclose(tmp1);
fclose(tmp2);
}

/***********************************************************************/
void GenerateSmartQueries(char *filename){
/* filename.prs contains a parsed set of queries of the form:

   w_1  w_2  ... w_n1
   w_1' w_2' ... w_n2'
   ...

   This function generates file filename.smr, containing
   a set of queries for Smart that are parseable with
   the generic Smart parser, and including all single-word 
   queries after each of the "big" queries. It also generates
   file filename.mst with all of the queries without the Smart
   separators, for later use... */

FILE *parsed, *smart, *master;

OpenFiles(filename, &parsed, &smart, &master);
ProcessInputFile(parsed, smart, master);
CloseFiles(parsed, smart, master);
  
}

/***********************************************************************/
void OpenFiles(char *filename, FILE **parsed, FILE **smart, FILE **master){
char f_parsed[FILENAME_MAX], f_smart[FILENAME_MAX], f_master[FILENAME_MAX];

/* Open input and output files */
sprintf(f_parsed, "%s/%s.prs", QUERY_DIR, filename);
sprintf(f_smart,  "%s/%s.smr", QUERY_DIR, filename);
sprintf(f_master, "%s/%s.mst", QUERY_DIR, filename);

if((*parsed=fopen(f_parsed, "r"))==(FILE *)NULL)
  Panic("Couldn't open f_parsed file");

if((*smart=fopen(f_smart, "w"))==(FILE *)NULL)
  Panic("Couldn't open f_smart file");

if((*master=fopen(f_master, "w"))==(FILE *)NULL)
  Panic("Couldn't open f_master file");
}

/***********************************************************************/
void ProcessInputFile(FILE *parsed, FILE *smart, FILE *master){
/* For each query in parsed, generate a Smart query into smart, and
   a plain text query into master.
   Also, for each query q in parsed, and each keyword in q, 
   generate one one-word query into smart and into master. */
char buffer[MAXLINE];

while(fgets(buffer, MAXLINE-1, parsed)!=NULL)
   /* process query (there's one query per line) */
   ProcessQuery(buffer, smart, master);
}

/***********************************************************************/
void ProcessQuery(char *buffer, FILE *smart, FILE *master){
/* buffer contains a single query. Process it and output the
   corresponding queries to files smart and master */

char word[MAXLINE];
char small_queries[MAXLINE];
char tmp_buffer[MAXLINE];
int num_words=0;

/* First, print the query separators and the entire query to smart */
fprintf(smart, ".I\n.W\n%s", buffer);

/* Then, parse the query and output each invidual keyword in it 
   as a separate query */

strcpy(tmp_buffer, buffer);
small_queries[0]='\0';

while(GetNextWord(&buffer, word)!=EOF){
  num_words++;
  strcat(small_queries, word);
  strcat(small_queries, "\n");
  fprintf(smart, ".I\n.W\n%s\n", word);
}

/* No more words in buffer */
/* Write info to master file, starting with the number of keywords in the 
   query. */
fprintf(master, "%d\n", num_words);
fprintf(master, "%s", tmp_buffer);
fprintf(master, "%s", small_queries);

}

/***********************************************************************/
int GetNextWord(char **buffer, char *word){
  int i=0;

  /* skip white space, and the line break if at the end of the query...*/
  while(((*buffer)[i]==' ')||((*buffer)[i]=='\n'))
    i++;

  if((*buffer)[i]=='\0')
    return EOF;

  /* There is a word to read */
  (*buffer)=(*buffer)+i;
  i=0;

  while(((*buffer)[i]!='\0') && ((*buffer)[i]!=' ') &&
        ((*buffer)[i]!='\n')){
    word[i]=(*buffer)[i];
    i++;
  }

  word[i]='\0';

  /* advance buffer pointer after the current word ... */
  (*buffer)=(*buffer)+i;

  return 0;
}

/***********************************************************************/
void GenerateQueryWeightVector(char *query, int num_words, int *w_vector){
/* Returns in array w_vector the query weight vector associated with
   query. Query has num_words words. The vector associated with
   query "a b c a a" will be "3 1 1 0 0," for example. */

char *terms;
int i,j;

/* load the num_words query terms into array terms
   and initialize w_vector */
terms=malloc(sizeof(char)*MAXLINE*num_words);

for(i=0; i<num_words; i++){
  w_vector[i]=-1;
  if(GetNextWord(&query, (terms+i*MAXLINE))==EOF)
    Panic("Can't parse query for generating query weights");
}

/* generate weights */
for(i=0; i<num_words; i++){
  if(w_vector[i]==-1){
    /* word not found earlier in the query: process */
    w_vector[i]=1;

    for(j=i+1; j<num_words; j++)
      /* check if word appears later in query */
      if(strcmp(terms+i*MAXLINE, terms+j*MAXLINE)==0){
        /* found word later */
        w_vector[i]++;
        w_vector[j]=0;
      }
  }
}
free(terms);
} 

/***********************************************************************/
void CloseFiles(FILE *parsed, FILE *smart, FILE *master){
fclose(parsed);
fclose(smart);
fclose(master);
}

/***********************************************************************/
void Panic(char *msg){
fprintf(stderr, "Problem: %s\n", msg);
exit(0);
}

/***********************************************************************/
void ProcessDBList(char *f_dblist){
/* Takes each directory dir (relative to the NEWS_DIR directory)
   listed in file dblist and 
   generates an entry (NEWS_DIR/dir, DB_ID(dir)) in a new file f_dblist.proc.
   DB_ID(dir) is the database id that will be associated with
   the database that dir specifies */

FILE *dblist, *processed_dblist;
char f_processed_dblist[FILENAME_MAX];
char dir[FILENAME_MAX], dbid[FILENAME_MAX];

if((dblist=fopen(f_dblist, "r"))==(FILE *)NULL)
  Panic("Couldn't open dblist file");

sprintf(f_processed_dblist, "%s.proc", f_dblist);
if((processed_dblist=fopen(f_processed_dblist, "w"))==(FILE *)NULL)
  Panic("Couldn't create processed_dblist");

while(fscanf(dblist, "%s", dir)==1){
  DB_ID(dir, dbid);
  fprintf(processed_dblist, "%s/%s %s\n", NEWS_DIR, dir, dbid);
}

fclose(dblist);
fclose(processed_dblist);
}

/***********************************************************************/
void DB_ID(char *db, char *dbid){
/* Generates a dbid from db, using  very simple character substitutions */
int i, len;

len=strlen(db);
for(i=0; i<len; i++){
  if(!isalnum(db[i]) && (db[i]!='-') && (db[i]!='_') && (db[i]!='.')) 
    dbid[i]='_';
  else
    dbid[i]=db[i];
}

dbid[i]='\0';
}

/***********************************************************************/
void ProcessQueryResults(char *dbid, long num_docs, 
      char *f_queries, char *res_dir, t_exp *experiment){

/* f_queries.mst is a file containing the original set of queries.
   res_dir is a directory containing the result of evaluating the
   queries in f_queries.mst on database dbid, which has
   num_docs in total.
   File RES_SUMMARY will contain the "summary" information
   (see below) for each query and database dbid.

   This function processes the query results and comes up with two
   "numbers" for each query: the first number captures the "goodness"
   of dbid for the query 
   and the second number is GlOSS's attempt at capturing this goodness.
   (These algorithms are specified in experiment.) */

FILE *queries, *results, *proc_results;
char rf_queries[FILENAME_MAX], f_results[FILENAME_MAX];

/* Generate the file names...*/
sprintf(rf_queries, "%s/%s.mst", QUERY_DIR, f_queries);
sprintf(f_results, "%s/%s.txt", res_dir, dbid);

OpenQueryFiles(rf_queries, f_results, RES_SUMMARY,
               &queries, &results, &proc_results);
ProcessQueryFiles(dbid, num_docs, queries, results, proc_results, experiment);
CloseQueryFiles(queries, results, proc_results, f_results);
}

/***********************************************************************/
void OpenQueryFiles(char *rf_queries, char *f_results, char *f_proc_results,
                    FILE **queries, FILE **results, FILE **proc_results){

char buffer[MAXLINE];

if((*queries=fopen(rf_queries, "r"))==(FILE *)NULL)
  Panic("Couldn't open queries file");

/* uncompress query-results file */
sprintf(buffer, "gunzip %s\n", f_results);
system(buffer); 

if((*results=fopen(f_results, "r"))==(FILE *)NULL)
  Panic("Couldn't open results file");

if((*proc_results=fopen(f_proc_results, "a"))==(FILE *)NULL)
  Panic("Couldn't create proc_results file");
}

/***********************************************************************/
void CloseQueryFiles(FILE *queries, FILE *results, FILE *proc_results,
                     char *f_results){

char buffer[MAXLINE];

fclose(queries);
fclose(results);
fclose(proc_results);

/* Compress the query-results file */
sprintf(buffer, "gzip %s\n", f_results);
system(buffer);
}

/***********************************************************************/
void ProcessQueryFiles(char *dbid, long num_docs, 
                       FILE *queries, FILE *results, FILE *proc_results,
                       t_exp *experiment){
/* For each query in queries, take the results for it and its individual
   one-keyword subqueries from results, and combine these results
   into a "real" goodness number, and a "GlOSS-predicted" goodness number,
   using the algorithms specified in experiment. */

int qid=1;  /* keeps track of the current query number */
            /* Smart numbers queries starting from 1 */

while(ProcessQueryData(&qid, dbid, num_docs, queries, results, proc_results, 
                       experiment)!=EOF);
     
}

/***********************************************************************/
int ProcessQueryData(int *qid, char *dbid, long num_docs, 
                     FILE *queries, FILE *results, FILE *proc_results,
                     t_exp *experiment){
/* Takes the next query (query number qid) from file queries and its
   associated one-keyword queries, and the corresponding results from
   file results, and generates two numbers, a "real" goodness number, 
   and a "GlOSS-predicted" goodness number, using the algorithms 
   described in experiment.

   Returns EOF if there are no more queries to process. */

int num_terms;
int result;
t_qres *query_results;

float real_good;
float gloss_good;

int *query_weights;

if((result=ReadQuery(queries, &num_terms, &query_weights))==EOF)
  return EOF;

/* Creates the query_results vector */
if((query_results=(t_qres *)malloc(sizeof(t_qres)*(num_terms+1)))==NULL)
   Panic("Malloc failed");

ReadAndCollapseResults(results, *qid, num_terms, query_results, experiment);
real_good =RealGoodness(num_terms, query_results, experiment);
gloss_good=GlOSSGoodness(num_terms, query_results, num_docs, experiment,
                         query_weights);

/* Gather statistics on whether the real and predicted goodnesses are zero */
if(real_good==0) zero_real_goodness++;
if(gloss_good==0) zero_gloss_goodness++;
total_goodness++;

/* Output format:

   <qid> <dbid> <num of docs with non-zero weight> <sum of the docs weights>
                <real "goodness"> <estimated "goodness">
 */

fprintf(proc_results, "%d\t%s %ld %f %f %f\n", 
   *qid, dbid, query_results[0].doc_counts, query_results[0].weight_sums, 
   real_good, gloss_good);

free(query_results);
free(query_weights);

(*qid)=(*qid)+num_terms+1;
return *qid;
}

/***********************************************************************/
int ReadQuery(FILE *queries, int *num_terms, int **query_weights){
/* Read next query and its subqueries from file queries.
   Returns in vector query_weights the weights associated with each 
   of the num_terms terms of the current query.
   Returns EOF if there are no more queries to process. */

char orig_query[MAXLINE];
char keyword[MAXLINE];
char buffer[MAXLINE];
int i;

/* First read number of keywords in the query */
if(fgets(buffer, MAXLINE, queries)==NULL)
  return EOF;

if(sscanf(buffer, "%d", num_terms)!=1)
  Panic("Query file in wrong format");

/* Now read the entire query */
if(fgets(orig_query, MAXLINE, queries)==NULL)
  Panic("Couldn't read original query from queries file");

/* And the num_terms individual keywords that compose the query */
for(i=0; i<(*num_terms); i++)
   if(fgets(keyword, MAXLINE, queries)==NULL)
     Panic("Couldn't read all keywords from queries file");

(*query_weights) = malloc(sizeof(int)*(*num_terms));
GenerateQueryWeightVector(orig_query, (*num_terms), (*query_weights));

return 0;
}

/***********************************************************************/
void ReadAndCollapseResults(FILE *results, int qid, int num_terms, 
                            t_qres *query_results, t_exp *experiment){

/* Reads the query results associated with query number qid and all of
   its num_terms subqueries from file results. 
   Then combines these results into vector query_results
   using the algorithms described in experiment. */

int i;
long count;
float sum;

for(i=0; i<=num_terms; i++){
   ReadAndCollapseResultsForOne(results, qid+i, &count, &sum, i, experiment); 
   query_results[i].doc_counts=count;
   query_results[i].weight_sums=sum;
}

return;
}

/***********************************************************************/
void ReadAndCollapseResultsForOne(FILE *results, int qid, 
            long *count, float *sum, int main_query, t_exp *experiment){
/* Reads the query results associated with query number qid 
   from file results. Then combines these results into count and sum, using
   the algorithms described in experiment. If main_query==0, then
   the current query is a "real" query, not a component of a real user query.
   We need this to collapse things according to the ranking algorithms
   described in experiment. */

int qnumber, aux1, aux2, aux3, aux4;
long dnumber;
float similarity;

(*count)=0;
(*sum)=0;

/* First check for "leftovers" in gbuffer. If gbuffer is empty,
   then read the new line in it. */  

while((gbuffer[0]!='\0')||(fgets(gbuffer, MAXLINE, results)!=NULL)){
  sscanf(gbuffer, "%d %ld %d %d %d %d %f", 
         &qnumber, &dnumber, &aux1, &aux2, &aux3, &aux4, &similarity);

  if(qnumber==qid){
    /* results correspond to the right query */

    if(main_query==0){
      /* For any ranking algorithm, we only "count" documents 
         in the answer for the "main" queries
         if they have a weight greater than the given threshold_real. */
      if(similarity>(experiment->threshold_real)){
        (*count)++;
        (*sum)=(*sum)+similarity;
      }
    }
    else{
      (*count)++;
      (*sum)=(*sum)+similarity;
    }

    /* empty gbuffer: we've already used its contents */
    gbuffer[0]='\0';
  }
  else
    /* read first line for a later query: leave it in the gbuffer... */
    break;
}
}

/***********************************************************************/
float RealGoodness(int num_terms, t_qres *query_results, t_exp *experiment){
/* Determines how "good" the current database is for the current query,
   based on the sum of the similarities between the db's documents and
   the query and the number of documents matching some of
   the query's keywords (given by query_results), and using the algorithm in 
   experiment */

if((experiment->real_rank_alg)==REAL_RANK_W)
  return query_results[0].weight_sums;
else
  /* real_rank_alg==REAL_RANK_D */
  return query_results[0].doc_counts;
}

/***********************************************************************/
float GlOSSGoodness(int num_terms, t_qres *query_results,
                    long num_docs, t_exp *experiment, int *query_weights){
/* Determines how "good" GlOSS would judge the current database for the
   current query to be, based on data for the individual keywords
   in the current query (given in query_results, with weights in 
   query_weights), and the algorithm in experiment. */

int i, j;
float goodness=0, added_weight=0;
long doc_count=0;
float weight_sum=0;

switch (experiment->gloss_rank_alg){
      
  case GLOSS_RANK_SUM_W:
  case GLOSS_RANK_SUM_D:
    /* GlOSS will assume that query words occur in documents with 
       negative correlation (Sum...) 
       GlOSS essentially assumes that no document contains 
       two query terms. */
    for(i=1; i<=num_terms; i++){
         if((query_results[i].doc_counts!=0)&&
            (query_weights[i-1]*
             (query_results[i].weight_sums/query_results[i].doc_counts) >
              experiment->threshold_gloss)){
           doc_count+=query_results[i].doc_counts;
           weight_sum+=query_results[i].weight_sums*query_weights[i-1];
         } 
    }

    if(experiment->gloss_rank_alg == GLOSS_RANK_SUM_W)
      goodness=weight_sum;
    else /* experiment->gloss_rank_alg == GLOSS_RANK_SUM_D */
      goodness=doc_count;

    break;  
      
  case GLOSS_RANK_MAX_W: 
  case GLOSS_RANK_MAX_D: 
    /* GlOSS will estimate the number of documents having weight > 
       threshold_gloss, using the high-correlation-of-keywords assumption. */ 

    qsort((void *)(query_results+1), (size_t)num_terms, 
         (size_t)(sizeof(t_qres)), (int (*)(void*,void*))query_results_compare);

    /* Now, find j (if it exists) such that:
       wj/fj + wj+1/fj+1 + ... + wn/fn  > threshold_gloss, but
               wj+1/fj+1 + ... + wn/fn <= threshold_gloss. 
     Then, under the assumptions, the only documents that will have
     weight > threshold_gloss are those containing (at least) terms 
     tj through tn. There are tj such documents. */

    added_weight=0;
    for(i=1; i<=num_terms; i++){
      if(query_results[i].doc_counts!=0)
        added_weight+=
            ((query_results[i].weight_sums)/(query_results[i].doc_counts));
    }
    
    if(added_weight<=(experiment->threshold_gloss)){
      /* no document (under the assumptions) has the necessary weight */
      doc_count=0;
      weight_sum=0;
    } 
    else{
      for(i=1; i<=num_terms; i++){
         if(query_results[i].doc_counts!=0)
           added_weight-=
               (query_results[i].weight_sums)/(query_results[i].doc_counts);
         if((added_weight<=(experiment->threshold_gloss))||(i==num_terms))
           break;
      }
      /* the weight sum of the documents that qualify is:
         w1+...+wj+ fj*(wj+1/fj+1 + ... + wn/fn) */
     
      doc_count=(query_results[i].doc_counts);  /* fj */
      weight_sum=doc_count*added_weight; /* fj*(wj+1/fj+1 + ... + wn/fn) */       
      for(j=1; j<=i; j++)
         weight_sum+=query_results[j].weight_sums;
    }
    
    if(experiment->gloss_rank_alg == GLOSS_RANK_MAX_W)
      goodness=weight_sum;
    else /* experiment->gloss_rank_alg == GLOSS_RANK_MAX_D */
      goodness=doc_count;

    break;

  default:
    Panic("Unknown GlOSS-rank algorithm");
    break;
} /* switch */

return goodness;
}

/***********************************************************************/
int query_results_compare(void *v1, void *v2)
{
 t_qres *t1, *t2;

 t1=(t_qres *)v1;
 t2=(t_qres *)v2;

 if((t1->doc_counts) < (t2->doc_counts))
   return -1;
 else{
   if((t1->doc_counts) == (t2->doc_counts))
     return 0;
   else
     return 1;
 }
}
 
/***********************************************************************/
void CompareRanks(int num_dbs, t_exp *experiment){
/* REAL_RANK contains, for each query, 
   the rank of databases according to their "goodness."
   GLOSS_RANK contains, for each query, 
   the rank of databases according to GlOSS predictions.
   experiment contains some parameters of the experiments.

   This function compares these two ranks query by query, 
   and prints statistics to the file specified in experiment. */

FILE *real, *gloss;

OpenRankFiles(REAL_RANK, GLOSS_RANK, &real, &gloss);
ComputeStats(num_dbs, real, gloss, experiment);
CloseRankFiles(real, gloss);
}

/***********************************************************************/
void OpenRankFiles(char *f_real, char *f_gloss, 
                   FILE **real, FILE **gloss){

if((*real=fopen(f_real, "r"))==(FILE *)NULL)
  Panic("Couldn't open f_real file");

if((*gloss=fopen(f_gloss, "r"))==(FILE *)NULL)
  Panic("Couldn't open f_gloss file");
}

/***********************************************************************/
void CloseRankFiles(FILE *real, FILE *gloss){

fclose(real);
fclose(gloss);
}

/***********************************************************************/
void ComputeStats(int num_dbs, FILE *real, FILE *gloss, t_exp *experiment){

int num_queries=0;

float *goodness_real, *goodness_gloss, *predicted_goodness;

float goodness_recall[MAXDBS], goodness_precision[MAXDBS];  

CreateVectors(num_dbs, &goodness_real, &goodness_gloss, &predicted_goodness,
              goodness_recall, goodness_precision);

while(ReadInOneQuery(num_dbs, real, gloss, 
             goodness_real, goodness_gloss, predicted_goodness)!=EOF){
   num_queries++;
   AnalyzeData(num_dbs, goodness_recall, goodness_precision, 
               goodness_real, goodness_gloss, predicted_goodness, experiment);
}

ReportResults(num_queries, num_dbs, goodness_recall, goodness_precision,
              experiment);
DestroyVectors(goodness_real, goodness_gloss, predicted_goodness);
}

/***********************************************************************/
void CreateVectors(int num_dbs, float **goodness_real, 
                   float **goodness_gloss, float **predicted_goodness,
                   float *goodness_recall, float *goodness_precision){
/* create vectors to keep the data on the two ranks for a single
   query */

int i;
int max_dbs;

*goodness_real=(float *)malloc(sizeof(float)*num_dbs);
*goodness_gloss=(float *)malloc(sizeof(float)*num_dbs);
*predicted_goodness=(float *)malloc(sizeof(float)*num_dbs);

/* initialize the statistics vectors */

if(num_dbs<MAXDBS)
  max_dbs=num_dbs;
else
  max_dbs=MAXDBS;

for(i=0; i<max_dbs; i++){
  goodness_recall[i]=0;
  goodness_precision[i]=0;
}
}

/***********************************************************************/
void DestroyVectors(float *goodness_real, float *goodness_gloss, 
                    float *predicted_goodness){

free(goodness_real);
free(goodness_gloss);
free(predicted_goodness);

}

/***********************************************************************/
int ReadInOneQuery(int num_dbs, FILE *real, FILE *gloss, 
                   float *goodness_real, float *goodness_gloss, 
                   float *predicted_goodness){

int qid, i;
char dbid[FILENAME_MAX];
long aux1;
float aux2, aux3;
char buffer[MAXLINE];

for(i=0; i<num_dbs; i++){

   if(fgets(buffer, MAXLINE, real)==NULL){
      if(i==0) 
        return EOF;
      else 
        Panic("Problem reading 'real' data");
   }

   if(sscanf(buffer, "%d %s %ld %f %f %f", &qid, dbid, 
                    &aux1, &aux2, &(goodness_real[i]), &aux3)!=6)
     Panic("Real rank in bad format");

   if((fgets(buffer, MAXLINE, gloss)==NULL)||
      (sscanf(buffer, "%d %s %ld %f %f %f", &qid, dbid, 
        &aux1, &aux2, &(goodness_gloss[i]), &(predicted_goodness[i]))!=6))
     Panic("Problem reading 'GlOSS' data");
}

return 0;
}

/***********************************************************************/
void AnalyzeData(int num_dbs, float *goodness_recall, float *goodness_precision,
                 float *goodness_real, float *goodness_gloss, 
                 float *predicted_goodness, t_exp *experiment){

/* Computes "recall" values at different cutting points: if we examine
   only the first n databases according to GlOSS's rank, 
   what do we get compared with what we would have gotten with
   the "best" n databases? Vector goodness_recall contains these recall
   values. Vector goodness_precision computes the fraction of useful
   databases among the first n databases in GlOSS's rank.  */

int n, i;
int max_dbs;
float ideal_goodness, gloss_goodness;
float new_recall, new_precision;
long chosen_dbs=0, useful_dbs=0; /* to compute precision_n: in the
                                    top n databases in the rank,
                                    there were chosen_dbs with
                                    non-zero estimate, and useful_dbs
                                    with non-zero goodness among them */

ideal_goodness=gloss_goodness=0;

if(num_dbs<MAXDBS)
  max_dbs=num_dbs;
else
  max_dbs=MAXDBS;

for(i=0; i<max_dbs; i++){
   ideal_goodness+=goodness_real[i];
   
   if(((experiment->prune)==NON_PRUNED_RANKS)||
      (predicted_goodness[i]>0)){
      /* will only add the real goodness of a database
         (to compute the overall goodness of the GlOSS rank)
         if that database was predicted to have non-zero goodness
         by GlOSS (PRUNED_RANKS option) */
      gloss_goodness+=goodness_gloss[i];
 
      chosen_dbs++;
      if(goodness_gloss[i]>0)
        useful_dbs++; 
   }
}

if(ideal_goodness==0)
  new_recall=1;
else{
  new_recall=gloss_goodness/ideal_goodness;
  if(new_recall>1) 
    Panic("Recall greater than one");
}

if(chosen_dbs==0)
  new_precision=1;
else{
  new_precision=useful_dbs/chosen_dbs;
  if(new_precision>1)
    Panic("Precision greater than one");
}

goodness_recall[max_dbs-1]+=new_recall;
goodness_precision[max_dbs-1]+=new_precision;

for(n=max_dbs-2; n>=0; n--){
  ideal_goodness-=goodness_real[n+1];

  if(((experiment->prune)==NON_PRUNED_RANKS)||
     (predicted_goodness[n+1]>0)){
    gloss_goodness-=goodness_gloss[n+1];

    chosen_dbs--;
    if(goodness_gloss[n+1]>0)
      useful_dbs--;
  }

  if(ideal_goodness==0)
    new_recall=1;
  else{
    new_recall=gloss_goodness/ideal_goodness;
    if(new_recall>1)
      Panic("Recall greater than one");
  }

  if(chosen_dbs==0)
    new_precision=1;
  else{
    new_precision=useful_dbs/chosen_dbs;
    if(new_precision>1)
      Panic("Precision greater than one");
  }
  goodness_recall[n]+=new_recall;
  goodness_precision[n]+=new_precision; 
}
}

/***********************************************************************/
void ReportResults(int num_queries, int num_dbs, 
                   float *goodness_recall, float *goodness_precision,
                   t_exp *experiment){

int i, max_dbs;
FILE *output;

if((output=fopen(experiment->output_file, "a"))==(FILE *)NULL)
   Panic("Can't open output file");

fprintf(output, "----------------------------------------------------------\n");
fprintf(output, "Queries:\t\t%d\n", num_queries);
fprintf(output, "Databases:\t\t%d\n", num_dbs);
fprintf(output, "Real_rank_algorithm:\t%d\n", experiment->real_rank_alg);
fprintf(output, "GlOSS_rank_algorithm:\t%d\n", experiment->gloss_rank_alg);
fprintf(output, "Threshold_Real:\t\t%f\n", experiment->threshold_real);
fprintf(output, "Threshold_GlOSS:\t%f\n", experiment->threshold_gloss);
if(experiment->prune == PRUNED_RANKS)
  /* No dbs with an estimated zero goodness are considered part of the rank */
  fprintf(output, "GlOSS_rank pruned\n\n");
else
  fprintf(output, "GlOSS_rank not pruned\n\n");

fprintf(output, "(1) Zero real goodness: \t %ld \n", zero_real_goodness);
fprintf(output, "(2) Zero GlOSS goodness:\t %ld \n", zero_gloss_goodness);
fprintf(output, "(3) Total goodness:\t\t %ld \n\n", total_goodness);
fprintf(output, "(1)/(3):\t %f \n", (zero_real_goodness*1.0)/total_goodness);
fprintf(output, "(2)/(3):\t %f \n\n", (zero_gloss_goodness*1.0)/total_goodness);

fprintf(output, "Num_DBs\t\tGoodness recall\tGoodness precision\n");

if(num_dbs<MAXDBS)
  max_dbs=num_dbs;
else
  max_dbs=MAXDBS;

for(i=0; i<max_dbs; i++)
   fprintf(output, "%d\t\t%f\t%f\n", i+1, goodness_recall[i]/num_queries,
                   goodness_precision[i]/num_queries); 

fprintf(output, "----------------------------------------------------------\n");
fclose(output);
}

/***********************************************************************/
void FeedSmart(char *proc_dblist, int from, int to, char *res_dir){
/* Takes each <db, dbid> pair in proc_dblist, and feeds it to the
   test_news... scripts, which generate the corresponding Smart database.
   Actually, process only pairs from (and including) 'from' to
   (and including) 'to.' Number dbs starting from 0. 

   After indexing each database, process the results from the binaries
   that Smart generated in directory res_dir, and generate the corresponding
   .txt files containing ASCII versions of the query results. */

FILE *tmp;
char db[FILENAME_MAX], dbid[FILENAME_MAX],
     bin_res[FILENAME_MAX], var_bin_res[FILENAME_MAX], 
     txt_res[FILENAME_MAX], txt_res_gz[FILENAME_MAX];
char local_docs[FILENAME_MAX];
char buffer[MAXLINE];
int count=0;

if((tmp=fopen(proc_dblist, "r"))==(FILE *)NULL)
  Panic("Couldn't open proc_dblist file");

while(fscanf(tmp, "%s %s", db, dbid)==2){
  if((count>=from)&&(count<=to)){

    /* First copy all newsgroup articles in a local directory,
       for efficiency and logging purposes... */
    
    sprintf(local_docs, "%s/%s", LOCAL_DIR, dbid);
    /* DEBUG */
    if(mkdir(local_docs, 0700)!=0)
      Panic("Can't create local directory for documents"); 
    if(chdir(db)!=0)
      Panic("Can't cd to newsgroup directory");

    sprintf(buffer, "cp * %s/.\n", local_docs);
    printf("DEBUG: %s\n", buffer);
    system(buffer);
    /* DEBUG */

    /* DEBUG 
    chdir(local_docs);
    system("gunzip *\n");
    DEBUG */

    /* Generate Smart indexes */
    sprintf(buffer, "%s %s %s\n", TEST_NTC_NNN, local_docs, dbid);
    printf("*********** %s\n", dbid);
    system(buffer);

    /* Extract text file with the query results for this database */
    sprintf(bin_res, "%s/%s", res_dir, dbid);
    sprintf(var_bin_res, "%s.var", bin_res);
    sprintf(txt_res, "%s.txt", bin_res);
    sprintf(txt_res_gz, "%s.gz", txt_res);

    /* Delete first old text query results */ 
    remove(txt_res_gz);

    sprintf(buffer, "%s tr_vec %s > %s\n", SMPRINT, bin_res, txt_res);
    system(buffer);

    /* Compress the query-results file */
    sprintf(buffer, "gzip %s\n", txt_res);
    system(buffer);

    /* Compress the documents in the local directory */
    if(chdir(local_docs)!=0)
      Panic("Can't cd to local directory for documents");
    system("gzip *\n");

    /* Remove the binary results files, for space saving... */
    remove(bin_res);
    remove(var_bin_res);
  }
  count++;
}
}

/***********************************************************************/
void GenerateRanks(char *proc_dblist, char *f_queries, char *res_dir,
                   int *num_dbs, t_exp *experiment){
/* First generates a single RES_SUMMARY file, with
   processed query results according to the ranking algorithms
   being used (specified in experiment).

   Then generates the "real" rank, in file REAL_RANK, and
   the GlOSS rank, in file GLOSS_RANK, using the data in RES_SUMMARY

   num_dbs returns the number of databases in proc_dblist. */

char db[FILENAME_MAX], dbid[FILENAME_MAX];
char buffer[MAXLINE];
FILE *tmp;

if((tmp=fopen(proc_dblist, "r"))==(FILE *)NULL)
  Panic("Couldn't open proc_dblist file");

/* clean up summary file */
remove(RES_SUMMARY);

(*num_dbs)=0;

/* process the data from each database separately */
while(fscanf(tmp, "%s %s", db, dbid)==2){
  (*num_dbs)++;
  /* Now process the data for this database... */
  ProcessQueryResults(dbid, 0, f_queries, res_dir, experiment);
}

sprintf(buffer, "%s \"%s\" \"%s\" \"%s\"\n", RANKER, 
        RES_SUMMARY, REAL_RANK, GLOSS_RANK);
system(buffer);
}

/***********************************************************************/
